#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021-2023, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Implementation of Velocity Auto-Correlation Function."""

# ============================================================================
# IMPORTS
# ============================================================================

import matplotlib.pyplot as plt

import numpy as np

import pandas as pd

from .core import MDObservable

# ============================================================================
# CLASSES
# ============================================================================


class VelocityAutocorrelationFunction(MDObservable):
    """Velocity Autocorrelation Function (VACF) implementation.

    The VACF is a measure of the correlation between the velocities of atoms
    in a molecular system as a function of time. It is calculated by
    multiplying the velocities of the atoms at a given time by the velocities
    of the same atoms at a reference time. The average over all atoms is
    considered.

    Parameters
    ----------
    frames : list
        a list with all the frames of the molecular dynamics trajectory, where
        each one is an `exma.core.AtomicSystem` and have the velocities
        defined.

    dt : int or float
        the timestep, how separated the measured frames are from each
        other, in the corresponding time units

    type_e : int or str, default="all"
        the type of the element for which the vacf is going to be calculated,
        by default it calculates the vacf of all atoms.

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames
    """

    def __init__(self, frames, dt, type_e="all", start=0, stop=-1, step=1):
        super().__init__(frames, start, stop, step)

        self.dt = dt
        self.type_e = type_e

    def _local_configure(self, frame):
        """Define the reference frame."""
        # mask of atoms of type e
        if self.type_e == "all":
            self.mask_e_ = np.full(frame.natoms, True)
        else:
            self.mask_e_ = frame._mask_type(self.type_e)

        # reference velocities
        self.vxref_ = frame.vx[self.mask_e_]
        self.vyref_ = frame.vy[self.mask_e_]
        self.vzref_ = frame.vz[self.mask_e_]

        self.velocity_autocorrelation_function = []

    def _accumulate(self, frame):
        """Calculate the vacf of a single frame."""
        vxsq = frame.vx[self.mask_e_] * self.vxref_
        vysq = frame.vy[self.mask_e_] * self.vyref_
        vzsq = frame.vz[self.mask_e_] * self.vzref_

        vacf = vxsq + vysq + vzsq

        self.velocity_autocorrelation_function.append(np.mean(vacf))

    def calculate(self):
        """Calculate the VACF.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` with the time and the vacf as columns.
        """
        super()._calculate()
        self.df_vacf_ = pd.DataFrame(
            {
                "t": self.dt * np.arange(0, len(self.frames), self.step),
                "vacf": np.array(self.velocity_autocorrelation_function),
            }
        )

        return self.df_vacf_

    def plot(self, ax=None, plot_kws=None):
        """Plot the calculated VACF.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axis, default=None
            current metplotlib axis

        plot_kws : dict, defualt=None
            additional keyword arguments that are passed and are documented
            in `matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.pyplot.Axis
            the axis with the plot
        """
        ax = plt.gca() if ax is None else ax

        plot_kws = {} if plot_kws is None else plot_kws

        ax.set_xlabel("t")
        ax.set_ylabel("vacf")
        ax.plot(self.df_vacf_["t"], self.df_vacf_["vacf"], **plot_kws)

        return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021-2023, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Implementation of functions of interest in electrochemistry."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

import pandas as pd

import scipy.interpolate

# ============================================================================
# FUNCTIONS
# ============================================================================


def fractional_volume_change(df, reference_atoms, reference_volume):
    r"""Fractional volume change (fvc) with respect to an element A.

    The *fvc* points are calculated using a normalization relative to
    the number of atoms in the reference structure and its volume

    .. math::
        fvc = {\frac{N_{ref}}{V_{ref}}} \left(
            {\frac{V_x}{N_x}} - {\frac{V_{ref}}{N_{ref}}}
        \right)

    where :math:`N_{ref}` and :math:`V_{ref}` are the number of atoms
    of type A and the volume of the reference structure, :math:`N_x` and
    :math:`V_x` are the number of atoms of type A and the volume of the
    structure at the :math:`x` concentration.

    Parameters
    ----------
    df : pd.DataFrame
        a `pd.DataFrame` with the values of `x` (concentrations) in
        the first column and then columns with the number of atoms of type A
        in each structure, the corresponding equilibrium values of the volume
        and, optionally, with its respective error.

    reference_atoms : int
        the number of atoms in the reference structure.

    reference_volume : float
        the volume, in the corresponding units, of the reference structue.

    Returns
    -------
    pd.DataFrame
        The input `pd.DataFrame` with extra columns with the fractional volume
        change and the respective error if it was possible to calculate, `fvc`
        and `err_fvc`, respectively.

    Raises
    ------
    KeyError
        if the number of atoms of type A or the volume are not defined in
        the `pd.DataFrame` as `"natoms_a"` and `"volume"`, respectively.

    Notes
    -----
    The names of the columns in the df must be the following ones:

    `"x"`: `x` values

    `"volume"`: volume

    `"natoms_a"`: number of atoms of type A

    `"err_volume"`: the error of each `"volume"` point, optional.
    """
    if ("natoms_a" not in df) or ("volume" not in df):
        raise KeyError(
            "The number of atoms of type A or the volume of the "
            "structures are not specified in the input DataFrame"
        )

    sfc = reference_volume / reference_atoms
    fvc = [
        ((volume / natoms) - sfc) / sfc
        for natoms, volume in zip(df.natoms_a, df.volume)
    ]

    df["fvc"] = np.array(fvc, dtype=np.float32)

    if "err_volume" in df:
        errfvc = df.err_volume / (df.natoms_a * sfc)
        df["err_fvc"] = np.array(errfvc, dtype=np.float32)

    return df


def formation_energy(df, reference_energy_a, reference_energy_b, fetype="x"):
    r"""Ideal approximation to the formation energy (FE).

    .. math::
        E_f(x) = E_x - (x \cdot E_b + E_a)

    where :math:`E_x` is the energy per atom of type A, :math:`x` the
    concentration and :math:`E_a` and :math:`E_b` the cohesive energies
    of A and B bulk materials.

    The formation energy relative to the end members can also be calculated

    .. math::
        E_f(x) = \frac{E - n_a \cdot E_a - n_b \cdot E_b}{n_a + n_b}

    Parameters
    ----------
    df : pd.DataFrame
        a `pd.DataFrame` with the values of `x` (concentrations) in
        the first column and then columns with the number of atoms of types
        A and the equilibrium values of the potential energy and, optionally,
        with its respective error.

    reference_energy_a : float
        the pure energy of element A in bulk

    reference_energy_b : float
        the pure energy of element B in bulk

    fetype : str, default="x"
        `"x"` or `"relative"`, where the first one is the formation energy as
        a function of concentrations and the second one is relative to end
        members

    Returns
    -------
    pd.DataFrame
        The input `pd.DataFrame` with the formation energy corresponding
        values, `"fe"`, and the respective error if it was possible to
        calculate, `"err_fe"`.

    Notes
    -----
    The names of the columns in the df must include:

    `"x"` o `"mol"`: `x` or `mol` fraction values

    `"natoms_a"`: number of atoms of type A

    `"natoms_b"`: number of atoms of type B

    `"epot"`: potential energy

    `"err_epot"`: the error of each `"epot"` point, optional.
    """
    if fetype == "x":
        if ("x" not in df) or ("natoms_a" not in df) or ("epot" not in df):
            raise KeyError(
                "The x values, the number of atoms of type A or the "
                "corresponding potential energies are not specified in "
                "the input DataFrame"
            )

        fe = [
            epot / na - (x * reference_energy_b + reference_energy_a)
            for x, na, epot in zip(df.x, df.natoms_a, df.epot)
        ]

    elif fetype == "relative":
        if (
            ("natoms_a" not in df)
            or ("natoms_b" not in df)
            or ("epot" not in df)
        ):
            raise KeyError(
                "The number of atoms of type A or B or the corresponding "
                "potential energies are not specified in the input DataFrame"
            )

        fe = [
            (epot - na * reference_energy_a - nb * reference_energy_b)
            / (na + nb)
            for na, nb, epot in zip(df.natoms_a, df.natoms_b, df.epot)
        ]

    df["fe"] = np.array(fe, dtype=np.float32)

    if "err_epot" in df:
        if fetype == "x":
            natoms = df.natoms_a
        elif fetype == "relative":
            natoms = df.natoms_a + df.natoms_b

        df["err_fe"] = np.asarray(df.err_epot / natoms, dtype=np.float32)

    return df


def voltage(df, nums=50, **kwargs):
    r"""Approximation to the voltage curve.

    The formation energies can be used as an approximation to the Gibbs
    formation energy, then the potential :math:`V` is given by:

    .. math::
        V(x) = -{\frac{dE_f(x)}{dx}}

    Parameters
    ----------
    df : pd.DataFrame
        a `pd.DataFrame` with the values of `x` (concentrations), the formation
        energy, `"fe"`, and, optionally, the respective error, `"err_fe"`.

    nums : int, default=50
        number of points at which to evaluate the spline and its
        derivative

    **kwargs
        additional keyword arguments that are passed and are documented
        in `scipy.interpolate.UnivariateSpline`

    Returns
    -------
    pd.DataFrame
        A `pd.DataFrame` with the value of the formation energy after
        the spline and the estimated value to the voltage in function of
        `x`.
    """
    weights = df.errfe if "err_fe" in df else None
    spline = scipy.interpolate.UnivariateSpline(
        df.x, df.fe, w=weights, **kwargs
    )
    dspline = spline.derivative()

    xpoints = np.linspace(np.min(df.x), np.max(df.x), nums)
    fe_spline = [spline(x) for x in xpoints]
    voltage = [-dspline(x) for x in xpoints]

    return pd.DataFrame(
        {
            "x": xpoints,
            "fe_spline": np.array(fe_spline, dtype=np.float32),
            "voltage": np.array(voltage, dtype=np.float32),
        }
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021-2023, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""Core classes of exma."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

# ============================================================================
# CLASSES
# ============================================================================


class AtomicSystem:
    """Define the configurations of the atomic system.

    Parameters
    ----------
    natoms : int, default=None
        the number of atoms in the atomic system

    box : np.array, default=None
        the box lenght in each direction

    types : np.array, default=None
        the types of the atoms

    idx : np.array, default=None
        the index of the atoms

    q : np.array, default=None
        a property of each atom in the atomic system

    x : np.array, default=None
        the positions of the atoms in the x direction

    y : np.array, default=None
        the positions of the atoms in the y direction

    z : np.array, default=None
        the positions of the atoms in the z direction

    ix : np.array, default=None
        the corresponding image of the positions of the atoms in the x
        direction

    iy : np.array, default=None
        the corresponding image of the positions of the atoms in the y
        direction

    iz : np.array, default=None
        the corresponding image of the positions of the atoms in the z
        direction

    vx : np.array, default=None
        the velocities of the atoms in the x direction

    vy : np.array, default=None
        the velocities of the atoms in the y direction

    vz : np.array, default=None
        the velocities of the atoms in the z direction
    """

    def __init__(
        self,
        natoms=None,
        box=None,
        types=None,
        idx=None,
        q=None,
        x=None,
        y=None,
        z=None,
        ix=None,
        iy=None,
        iz=None,
        vx=None,
        vy=None,
        vz=None,
    ):
        self.natoms = natoms
        self.box = box

        self.idx = idx
        self.types = types

        self.q = q

        self.x = x
        self.y = y
        self.z = z

        self.ix = ix
        self.iy = iy
        self.iz = iz

        self.vx = vx
        self.vy = vy
        self.vz = vz

    def _mask_type(self, atom_type):
        """Get a masked array by an specific type of atom."""
        return self.types == atom_type

    def _natoms_type(self, mask_type):
        """Count the number of atoms of an specific type."""
        return np.count_nonzero(mask_type)

    def _sorted(self):
        """Tells if the array x is sorted (-> True) or not (-> False).

        Note that for xyz files self.idx is None and the frame is sorted by
        definition.
        """
        return (np.diff(self.idx) >= 0).all() if self.idx is not None else True

    def _sort(self, dontsort=("natoms", "box")):
        """Sort the Atomic System from the sortening of the atoms id."""
        id_argsort = np.argsort(self.idx)

        for key in self.__dict__.keys():
            if (
                key.startswith("_")
                or key in dontsort
                or self.__dict__[key] is None
            ):
                continue

            self.__dict__[key] = self.__dict__[key][id_argsort]

        return self

    def _unwrap(self, m=None):
        """Unwrap the Atomic System m masked outside the box."""
        m = m if m is not None else np.full(self.natoms, True)

        self.x[m] = self.x[m] + self.box[0] * self.ix[m]
        self.y[m] = self.y[m] + self.box[1] * self.iy[m]
        self.z[m] = self.z[m] + self.box[2] * self.iz[m]

        return self

    def _wrap(self, m=None):
        """Wrap the Atomic System m masked inside the box."""
        m = m if m is not None else np.full(self.natoms, True)

        self.x[m] = np.mod(self.x[m], self.box[0])
        self.y[m] = np.mod(self.y[m], self.box[1])
        self.z[m] = np.mod(self.z[m], self.box[2])

        return self


class MDObservable:
    """Class to define the structure of the molecular dynamics observable.

    Parameters
    ----------
    frames : list
        a list with all the frames of the molecular dynamics trajectory, where
        each one is an `exma.core.AtomicSystem`.

    start : int, default=0
        the initial frame

    stop : int, default=-1
        the last frame, by default -1 means the last

    step : int, default=1
        the incrementation if it is necessary to skip frames
    """

    def __init__(self, frames, start=0, stop=-1, step=1):
        self.frames = frames

        self.start = start
        self.stop = stop
        self.step = step

    def _local_configure(self, frame):
        """Specific configuration of each observable."""
        raise NotImplementedError("Implemented in child classes.")

    def _accumulate(self, frame):
        """Accumulate the data of the frame."""
        raise NotImplementedError("Implemented in child classes.")

    def _end(self, frame):
        """Finish the calculation and normilize the data."""
        raise NotImplementedError("Implemented in child classes.")

    def _calculate(self, box=None):
        """Observable main loop, leave everything ready to do self._end().

        Parameters
        ----------
        box : np.array, default=None
            the lenght of the box in each x, y, z direction, required when
            the trajectory comes from an xyz file.
        """
        # select the production frames
        self.frames = (
            self.frames[self.start :: self.step]
            if self.stop == -1
            else self.frames[self.start : self.stop : self.step]
        )

        for i, frame in enumerate(self.frames):
            # add the box if not in frame
            frame.box = box if box is not None else frame.box

            # sort the frames if is not sorted, this might not be necessary
            # for all observables and all trajectories
            frame = frame._sort() if not frame._sorted() else frame

            if i == 0:
                self._local_configure(frame)

            self._accumulate(frame)

    def calculate(self):
        """Calculate the observable."""
        raise NotImplementedError("Implemented in child classes.")

    def to_dataframe(self):
        """Convert the results to pandas.DataFrame."""
        raise NotImplementedError("Implemented in child classes.")

    def plot(self):
        """Make a plot of the observable."""
        raise NotImplementedError("Implemented in child classes.")


class TrajectoryReader:
    """Class to read trajectory files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories in xyz format are

    ftype : str
        different type of files depending on the child class.
    """

    def __init__(self, filename, ftype):
        self.filename = filename
        self.ftype = ftype

    def __enter__(self):
        """Use the open() method."""
        self.file_traj_ = open(self.filename, "r")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Use the close() method."""
        self.file_traj_.close()

    def read_frame(self):
        """Read the actual frame of the file."""
        raise NotImplementedError("Implemented in child classes.")

    def read_traj(self):
        """Read all the trajectory of the file."""
        traj = []
        try:
            while True:
                traj.append(self.read_frame())
        except (EOFError, NotImplementedError):
            ...

        return traj


class TrajectoryWriter:
    """Class to write trajectory files.

    Parameters
    ----------
    filename : str
        name of the file where the trajectories in xyz format are going to
        be written

    ftype : str
        different type of files depending on the child class.
    """

    def __init__(self, filename, ftype):
        self.filename = filename
        self.ftype = ftype

    def __enter__(self):
        """Use the open() method."""
        self.file_traj_ = open(self.filename, "w")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Use the close() method."""
        self.file_traj_.close()

    def write_frame(self, frame):
        """Write the actual frame on the file."""
        raise NotImplementedError("Implemented in child classes.")

    def write_traj(self, frames):
        """Write all frames of the trajectory to a file."""
        for frame in frames:
            self.write_frame(frame)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of exma (https://github.com/fernandezfran/exma/).
# Copyright (c) 2021-2023, Francisco Fernandez
# License: MIT
#   Full Text: https://github.com/fernandezfran/exma/blob/master/LICENSE

# ============================================================================
# DOCS
# ============================================================================

"""exma IO module for interacting with molecular dynamics files."""

# ============================================================================
# IMPORTS
# ============================================================================

import numpy as np

from . import reader
from . import writer

# ============================================================================
# FUNCTIONS
# ============================================================================


def xyz2lammpstrj(xyztraj, lammpstrj_name, cell_info, xyzftype="xyz"):
    """Rewrite an xyz file to a lammpstrj file.

    Parameters
    ----------
    xyztraj : str
        the name of the file with the xyz trajectory.

    lammpstrj_name : str
        the name of the file with the lammpstrj trajectory.

    cell_info : dict
        with the `box`, the lenght of the box in each direction, another
        dictionary identified with the `type` key that has within it a
        correspondence between the elements present in xyz file with integer
        identification numbers, e.g. {"Sn": 1, "O": 2}

    xyzftype : str, default="xyz"
        the `ftype` of xyz file.
    """
    frames = reader.read_xyz(xyztraj, xyzftype)

    for frame in frames:
        frame.box = cell_info["box"]
        frame.idx = np.arange(1, frame.natoms + 1)
        frame.types = [cell_info["type"][t] for t in frame.types]

    writer.write_lammpstrj(frames, lammpstrj_name)


def xyz2inlmp(xyztraj, inlammps_name, cell_info, nframe=-1, xyzftype="xyz"):
    """Write an xyz frame to an input data file of LAMMPS.

    Parameters
    ----------
    xyztraj : str
        the name of the file with the xyz trajectory.

    inlammps_name : str
        the name of the file to write to.

    cell_info : dict
        with the `box`, the lenght of the box in each direction, another
        dictionary identified with the `type` key that has within it a
        correspondence between the elements present in xyz file with integer
        identification numbers, e.g. {"Sn": 1, "O": 2}

    nframe : int, default=-1
        number of the frame to write, by default is -1, that is, the last.

    xyzftype : str, default="xyz"
        the `ftype` of xyz file.

    Raises
    ------
    IndexError
        If the number of the frame to write is not in the trajectory file.
    """
    frames = reader.read_xyz(xyztraj, xyzftype)
    try:
        frame = frames[nframe]
    except IndexError as exc:
        errmsg = f"frame {nframe} does not exist in the trajectory file."
        raise IndexError(errmsg) from exc

    frame.box = cell_info["box"]
    frame.idx = np.arange(1, frame.natoms + 1)
    frame.types = [cell_info["type"][t] for t in frame.types]
    frame.q = cell_info["q"] if "q" in cell_info.keys() else None

    writer.write_in_lammps(frame, inlammps_name)


def lammpstrj2xyz(lammpstrjtraj, xyz_name, type_info, xyzftype="xyz"):
    """Rewrite a lammpstrj file to an xyz file.

    Parameters
    ----------
    lammpstrjtraj : str
        the name of the file with the lammpstrj trajectory.

    xyz_name : str
        the name of the file with the lammpstrj trajectory.

    type_info : dict
        a correspondence between the elements id present in lammpstrj file
        with str element, e.g. {1: "Sn", 2: "O"}

    xyzftype : str, default="xyz"
        the `ftype` of xyz file.
    """
    frames = reader.read_lammpstrj(lammpstrjtraj)

    for frame in frames:
        frame = frame._sort() if not frame._sorted() else frame
        frame.types = [type_info[t] for t in frame.types]

    writer.write_xyz(frames, xyz_name, xyzftype)


def lammpstrj2inlmp(lammpstrjtraj, inlammps_name, nframe=-1):
    """Write a lammpstrj frame to an input data file of LAMMPS.

    Parameters
    ----------
    lammpstrjtraj : str
        the name of the file with the lammpstrj trajectory.

    inlammps_name : str
        the name of the file to write to.

    nframe : int, default=-1
        number of the frame to write, by default is -1, that is, the last.

    Raises
    ------
    IndexError
        If the number of the frame to write is not in the trajectory file.
    """
    frames = reader.read_lammpstrj(lammpstrjtraj)
    try:
        frame = frames[nframe]
    except IndexError as exc:
        errmsg = f"frame {nframe} does not exist in the trajectory file."
        raise IndexError(errmsg) from exc

    frame = frame._sort() if not frame._sorted() else frame

    writer.write_in_lammps(frame, inlammps_name)
